@@ -7,6 +7,8 @@
 // Various tools for parsing LLVM. Utilities of various sorts, that are
 // specific to Emscripten (and hence not in utility.js).
 
+const FOUR_GB = 4 * 1024 * 1024 * 1024;
+
 let currentlyParsedFilename = '';
 
 // Does simple 'macro' substitution, using Django-like syntax,
@@ -31,43 +33,69 @@ function processMacros(text) {
 // Param filenameHint can be passed as a description to identify the file that is being processed, used
 // to locate errors for reporting and for html files to stop expansion between <style> and </style>.
 function preprocess(text, filenameHint) {
+  if (EXPORT_ES6 && USE_ES6_IMPORT_META) {
+    // `eval`, Terser and Closure don't support module syntax; to allow it,
+    // we need to temporarily replace `import.meta` usages with placeholders
+    // during preprocess phase, and back after all the other ops.
+    // See also: `phase_final_emitting` in emcc.py.
+    text = text.replace(/\bimport\.meta\b/g, 'EMSCRIPTEN$IMPORT$META');
+  }
+
+  const IGNORE = 0;
+  const SHOW = 1;
+  // This state is entered after we have shown one of the block of an if/elif/else sequence.
+  // Once we enter this state we dont show any blocks or evaluate any
+  // conditions until the sequence ends.
+  const IGNORE_ALL = 2;
+  const showStack = [];
+  const showCurrentLine = () => showStack.every((x) => x == SHOW);
+
   currentlyParsedFilename = filenameHint;
+  const fileExt = (filenameHint) ? filenameHint.split('.').pop().toLowerCase() : '';
+  const isHtml = (fileExt === 'html' || fileExt === 'htm') ? true : false;
+  let inStyle = false;
+  const lines = text.split('\n');
+  let ret = '';
+  let emptyLine = false;
+
   try {
-    const fileExt = (filenameHint) ? filenameHint.split('.').pop().toLowerCase() : '';
-    const isHtml = (fileExt === 'html' || fileExt === 'htm') ? true : false;
-    let inStyle = false;
-    const lines = text.split('\n');
-    let ret = '';
-    const showStack = [];
-    let emptyLine = false;
     for (let i = 0; i < lines.length; i++) {
       let line = lines[i];
       try {
         if (line[line.length - 1] === '\r') {
           line = line.substr(0, line.length - 1); // Windows will have '\r' left over from splitting over '\r\n'
         }
-        if (isHtml && line.indexOf('<style') !== -1 && !inStyle) {
+        if (isHtml && line.includes('<style') && !inStyle) {
           inStyle = true;
         }
-        if (isHtml && line.indexOf('</style') !== -1 && inStyle) {
+        if (isHtml && line.includes('</style') && inStyle) {
           inStyle = false;
         }
 
         if (!inStyle) {
           const trimmed = line.trim();
-          if (trimmed[0] === '#') {
+          if (trimmed.startsWith('#')) {
             const first = trimmed.split(' ', 1)[0];
-            if (first == '#if' || first == '#ifdef') {
+            if (first == '#if' || first == '#ifdef' || first == '#elif') {
               if (first == '#ifdef') {
                 warn('warning: use of #ifdef in js library.  Use #if instead.');
               }
+              if (first == '#elif') {
+                const curr = showStack.pop();
+                if (curr == SHOW || curr == IGNORE_ALL) {
+                  // If we showed to previous block we enter the IGNORE_ALL state
+                  // and stay there until endif is seen
+                  showStack.push(IGNORE_ALL);
+                  continue;
+                }
+              }
               const after = trimmed.substring(trimmed.indexOf(' '));
               const truthy = !!eval(after);
-              showStack.push(truthy);
+              showStack.push(truthy ? SHOW : IGNORE);
             } else if (first === '#include') {
-              if (showStack.indexOf(false) === -1) {
+              if (showCurrentLine()) {
                 let filename = line.substr(line.indexOf(' ') + 1);
-                if (filename.indexOf('"') === 0) {
+                if (filename.startsWith('"')) {
                   filename = filename.substr(1, filename.length - 2);
                 }
                 const included = read(filename);
@@ -80,7 +108,12 @@ function preprocess(text, filenameHint) {
               }
             } else if (first === '#else') {
               assert(showStack.length > 0);
-              showStack.push(!showStack.pop());
+              const curr = showStack.pop();
+              if (curr == IGNORE) {
+                showStack.push(SHOW);
+              } else {
+                showStack.push(IGNORE);
+              }
             } else if (first === '#endif') {
               assert(showStack.length > 0);
               showStack.pop();
@@ -88,7 +121,7 @@ function preprocess(text, filenameHint) {
               throw new Error(`Unknown preprocessor directive on line ${i}: ``${line}```);
             }
           } else {
-            if (showStack.indexOf(false) === -1) {
+            if (showCurrentLine()) {
               // Never emit more than one empty line at a time.
               if (emptyLine && !line) {
                 continue;
@@ -102,7 +135,7 @@ function preprocess(text, filenameHint) {
             }
           }
         } else { // !inStyle
-          if (showStack.indexOf(false) === -1) {
+          if (showCurrentLine()) {
             ret += line + '\n';
           }
         }
@@ -462,7 +495,7 @@ function ensureDot(value) {
   // if already dotted, or Infinity or NaN, nothing to do here
   // if smaller than 1 and running js opts, we always need to force a coercion
   // (0.001 will turn into 1e-3, which has no .)
-  if ((value.indexOf('.') >= 0 || /[IN]/.test(value))) return value;
+  if ((value.includes('.') || /[IN]/.test(value))) return value;
   const e = value.indexOf('e');
   if (e < 0) return value + '.0';
   return value.substr(0, e) + '.0' + value.substr(e);
@@ -856,25 +889,17 @@ function getHeapForType(type, unsigned) {
   assert(false, 'bad heap type: ' + type);
 }
 
-function makeGetTempRet0() {
-  return '(getTempRet0() | 0)';
-}
-
-function makeSetTempRet0(value) {
-  return 'setTempRet0((' + value + ') | 0)';
-}
-
 // Takes a pair of return values, stashes one in tempRet0 and returns the other.
 // Should probably be renamed to `makeReturn64` but keeping this old name in
 // case external JS library code uses this name.
 function makeStructuralReturn(values) {
   assert(values.length == 2);
-  return makeSetTempRet0(values[1]) + '; return ' + asmCoercion(values[0], 'i32');
+  return 'setTempRet0(' + values[1] + '); return ' + asmCoercion(values[0], 'i32');
 }
 
 function makeThrow(what) {
-  if (ASSERTIONS && DISABLE_EXCEPTION_CATCHING == 1) {
-    what += ' + " - Exception catching is disabled, this exception cannot be caught. Compile with -s DISABLE_EXCEPTION_CATCHING=0 or DISABLE_EXCEPTION_CATCHING=2 to catch."';
+  if (ASSERTIONS && DISABLE_EXCEPTION_CATCHING) {
+    what += ' + " - Exception catching is disabled, this exception cannot be caught. Compile with -s NO_DISABLE_EXCEPTION_CATCHING or -s EXCEPTION_CATCHING_ALLOWED=[..] to catch."';
     if (MAIN_MODULE) {
       what += ' + " (note: in dynamic linking, if a side module wants exceptions, the main module must be built with that support)"';
     }
@@ -992,7 +1017,7 @@ function asmFFICoercion(value, type) {
 }
 
 function makeDynCall(sig, funcPtr) {
-  assert(sig.indexOf('j') == -1, 'Cannot specify 64-bit signatures ("j" in signature string) with makeDynCall!');
+  assert(!sig.includes('j'), 'Cannot specify 64-bit signatures ("j" in signature string) with makeDynCall!');
 
   const returnExpr = (sig[0] == 'v') ? '' : 'return';
 
@@ -1053,6 +1078,8 @@ function addAtInit(code) {
   ATINITS.push(code);
 }
 
+// TODO(sbc): There are no more uses to ATMAINS or addAtMain in emscripten.
+// We should look into removing these.
 global.ATMAINS = [];
 
 function addAtMain(code) {
@@ -1092,13 +1119,6 @@ function makeRetainedCompilerSettings() {
 // In wasm, the heap size must be a multiple of 64KiB.
 const WASM_PAGE_SIZE = 65536;
 
-// Page size reported by some POSIX calls, mostly filesystem. This does not
-// depend on the memory page size which differs between wasm and asm.js, and
-// makes us report a consistent value despite the compile target. However,
-// perhaps we should unify all the page sizes (especially after fastcomp is
-// gone TODO).
-const POSIX_PAGE_SIZE = 16384;
-
 // Receives a function as text, and a function that constructs a modified
 // function, to which we pass the parsed-out name, arguments, and body of the
 // function. Returns the output of that function.
@@ -1194,7 +1214,7 @@ function makeModuleReceiveWithVar(localName, moduleName, defaultValue, noAssert)
 function makeRemovedFSAssert(fsName) {
   if (!ASSERTIONS) return;
   const lower = fsName.toLowerCase();
-  if (SYSTEM_JS_LIBRARIES.indexOf('library_' + lower + '.js') >= 0) return '';
+  if (SYSTEM_JS_LIBRARIES.includes('library_' + lower + '.js')) return '';
   return `var ${fsName} = '${fsName} is no longer included by default; build with -l${lower}.js';`;
 }
 
@@ -1219,11 +1239,20 @@ function makeAsmImportsAccessInPthread(variable) {
     return `imports['${variable}']`;
   }
   // In non-MODULARIZE builds, can access the imports from global scope.
-  return variable;
+  return `self.${variable}`;
+}
+
+function _asmjsDemangle(symbol) {
+  if (symbol in WASM_SYSTEM_EXPORTS) {
+    return symbol;
+  }
+  // Strip leading "_"
+  assert(symbol.startsWith('_'));
+  return symbol.substr(1);
 }
 
 function hasExportedFunction(func) {
-  return Object.keys(EXPORTED_FUNCTIONS).indexOf(func) != -1;
+  return Object.keys(WASM_EXPORTS).includes(_asmjsDemangle(func));
 }
 
 // JS API I64 param handling: if we have BigInt support, the ABI is simple,
@@ -1278,7 +1307,7 @@ function addReadyPromiseAssertions(promise) {
 }
 
 function makeMalloc(source, param) {
-  if ('_malloc' in IMPLEMENTED_FUNCTIONS) {
+  if (hasExportedFunction('_malloc')) {
     return `_malloc(${param})`;
   }
   // It should be impossible to call some functions without malloc being
@@ -1291,3 +1320,21 @@ function makeMalloc(source, param) {
   }
   return `abort('malloc was not included, but is needed in ${source}. Adding "_malloc" to EXPORTED_FUNCTIONS should fix that. This may be a bug in the compiler, please file an issue.');`;
 }
+
+// Adds a call to runtimeKeepalivePush, if needed by the current build
+// configuration.
+// We skip this completely in MINIMAL_RUNTIME and also in builds that
+// don't ever need to exit the runtime.
+function runtimeKeepalivePush() {
+  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && USE_PTHREADS == 0)) return '';
+  return 'runtimeKeepalivePush();';
+}
+
+// Adds a call to runtimeKeepalivePush, if needed by the current build
+// configuration.
+// We skip this completely in MINIMAL_RUNTIME and also in builds that
+// don't ever need to exit the runtime.
+function runtimeKeepalivePop() {
+  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && USE_PTHREADS == 0)) return '';
+  return 'runtimeKeepalivePop();';
+}
