@@ -15,7 +15,9 @@
 #include <math.h>
 #include <poll.h>
 #include <pthread.h>
+#include <stdatomic.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/ioctl.h>
@@ -25,6 +27,7 @@
 #include <sys/statvfs.h>
 #include <sys/time.h>
 #include <termios.h>
+#include <threads.h>
 #include <unistd.h>
 #include <utime.h>
 
@@ -85,6 +88,19 @@ int sched_get_priority_min(int policy) {
     return 0;
 }
 
+int pthread_mutexattr_getprioceiling(const pthread_mutexattr_t *restrict attr, int *restrict prioceiling)
+{
+  // Not supported either in Emscripten or musl, return a faked value.
+  if (prioceiling) *prioceiling = 99;
+  return 0;
+}
+
+int pthread_mutexattr_setprioceiling(pthread_mutexattr_t *attr, int prioceiling)
+{
+  // Not supported either in Emscripten or musl, return an error.
+  return EPERM;
+}
+
 int pthread_setcancelstate(int new, int* old) {
   if (new > 1U)
     return EINVAL;
@@ -154,20 +170,6 @@ void emscripten_thread_sleep(double msecs) {
     EM_THREAD_STATUS_SLEEPING, EM_THREAD_STATUS_RUNNING);
 }
 
-int nanosleep(const struct timespec* req, struct timespec* rem) {
-  if (!req || req->tv_nsec < 0 || req->tv_nsec > 999999999L || req->tv_sec < 0) {
-    errno = EINVAL;
-    return -1;
-  }
-  emscripten_thread_sleep(req->tv_sec * 1000.0 + req->tv_nsec / 1e6);
-  return 0;
-}
-
-int usleep(unsigned usec) {
-  emscripten_thread_sleep(usec / 1e3);
-  return 0;
-}
-
 // Allocator and deallocator for em_queued_call objects.
 static em_queued_call* em_queued_call_malloc() {
   em_queued_call* call = (em_queued_call*)malloc(sizeof(em_queued_call));
@@ -179,13 +181,17 @@ static em_queued_call* em_queued_call_malloc() {
   }
   return call;
 }
+
 static void em_queued_call_free(em_queued_call* call) {
   if (call)
     free(call->satelliteData);
   free(call);
 }
 
-void emscripten_async_waitable_close(em_queued_call* call) { em_queued_call_free(call); }
+void emscripten_async_waitable_close(em_queued_call* call) {
+  assert(call->operationDone);
+  em_queued_call_free(call);
+}
 
 extern double emscripten_receive_on_main_thread_js(int functionIndex, int numCallArgs, double* args);
 extern int _emscripten_notify_thread_queue(pthread_t targetThreadId, pthread_t mainThreadId);
@@ -352,7 +358,7 @@ static void _do_call(em_queued_call* q) {
   // If the caller is detached from this operation, it is the main thread's responsibility to free
   // up the call object.
   if (q->calleeDelete) {
-    emscripten_async_waitable_close(q);
+    em_queued_call_free(q);
     // No need to wake a listener, nothing is listening to this since the call object is detached.
   } else {
     // The caller owns this call object, it is listening to it and will free it up.
@@ -363,10 +369,11 @@ static void _do_call(em_queued_call* q) {
 
 #define CALL_QUEUE_SIZE 128
 
+// Shared data synchronized by call_queue_lock.
 typedef struct CallQueue {
   void* target_thread;
   em_queued_call** call_queue;
-  int call_queue_head; // Shared data synchronized by call_queue_lock.
+  int call_queue_head;
   int call_queue_tail;
   struct CallQueue* next;
 } CallQueue;
@@ -377,9 +384,8 @@ typedef struct CallQueue {
 static pthread_mutex_t call_queue_lock = PTHREAD_MUTEX_INITIALIZER;
 static CallQueue* callQueue_head = 0;
 
-static CallQueue* GetQueue(
-  void* target) // Not thread safe, call while having call_queue_lock obtained.
-{
+// Not thread safe, call while having call_queue_lock obtained.
+static CallQueue* GetQueue(void* target) {
   assert(target);
   CallQueue* q = callQueue_head;
   while (q && q->target_thread != target)
@@ -387,9 +393,8 @@ static CallQueue* GetQueue(
   return q;
 }
 
-static CallQueue* GetOrAllocateQueue(
-  void* target) // Not thread safe, call while having call_queue_lock obtained.
-{
+// Not thread safe, call while having call_queue_lock obtained.
+static CallQueue* GetOrAllocateQueue(void* target) {
   CallQueue* q = GetQueue(target);
   if (q)
     return q;
@@ -414,14 +419,14 @@ static CallQueue* GetOrAllocateQueue(
 EMSCRIPTEN_RESULT emscripten_wait_for_call_v(em_queued_call* call, double timeoutMSecs) {
   int r;
 
-  int done = emscripten_atomic_load_u32(&call->operationDone);
+  int done = atomic_load(&call->operationDone);
   if (!done) {
     double now = emscripten_get_now();
     double waitEndTime = now + timeoutMSecs;
     emscripten_set_current_thread_status(EM_THREAD_STATUS_WAITPROXY);
     while (!done && now < waitEndTime) {
       r = emscripten_futex_wait(&call->operationDone, 0, waitEndTime - now);
-      done = emscripten_atomic_load_u32(&call->operationDone);
+      done = atomic_load(&call->operationDone);
       now = emscripten_get_now();
     }
     emscripten_set_current_thread_status(EM_THREAD_STATUS_RUNNING);
@@ -442,17 +447,16 @@ EMSCRIPTEN_RESULT emscripten_wait_for_call_i(
 
 static pthread_t main_browser_thread_id_ = 0;
 
-void EMSCRIPTEN_KEEPALIVE emscripten_register_main_browser_thread_id(
+void emscripten_register_main_browser_thread_id(
   pthread_t main_browser_thread_id) {
   main_browser_thread_id_ = main_browser_thread_id;
 }
 
-pthread_t EMSCRIPTEN_KEEPALIVE emscripten_main_browser_thread_id() {
+pthread_t emscripten_main_browser_thread_id() {
   return main_browser_thread_id_;
 }
 
-int EMSCRIPTEN_KEEPALIVE do_emscripten_dispatch_to_thread(
-  pthread_t target_thread, em_queued_call* call) {
+int _emscripten_do_dispatch_to_thread(pthread_t target_thread, em_queued_call* call) {
   assert(call);
 
   // #if PTHREADS_DEBUG // TODO: Create a debug version of pthreads library
@@ -481,8 +485,8 @@ int EMSCRIPTEN_KEEPALIVE do_emscripten_dispatch_to_thread(
     q->call_queue = malloc(
       sizeof(em_queued_call*) * CALL_QUEUE_SIZE); // Shared data synchronized by call_queue_lock.
 
-  int head = emscripten_atomic_load_u32((void*)&q->call_queue_head);
-  int tail = emscripten_atomic_load_u32((void*)&q->call_queue_tail);
+  int head = q->call_queue_head;
+  int tail = q->call_queue_tail;
   int new_tail = (tail + 1) % CALL_QUEUE_SIZE;
 
   while (new_tail == head) { // Queue is full?
@@ -493,8 +497,8 @@ int EMSCRIPTEN_KEEPALIVE do_emscripten_dispatch_to_thread(
     if (target_thread == emscripten_main_browser_thread_id()) {
       emscripten_futex_wait((void*)&q->call_queue_head, head, INFINITY);
       pthread_mutex_lock(&call_queue_lock);
-      head = emscripten_atomic_load_u32((void*)&q->call_queue_head);
-      tail = emscripten_atomic_load_u32((void*)&q->call_queue_tail);
+      head = q->call_queue_head;
+      tail = q->call_queue_tail;
       new_tail = (tail + 1) % CALL_QUEUE_SIZE;
     } else {
       // For the queues of other threads, just drop the message.
@@ -522,32 +526,30 @@ int EMSCRIPTEN_KEEPALIVE do_emscripten_dispatch_to_thread(
     }
   }
 
-  emscripten_atomic_store_u32((void*)&q->call_queue_tail, new_tail);
-
+  q->call_queue_tail = new_tail;
   pthread_mutex_unlock(&call_queue_lock);
-
   return 0;
 }
 
-void EMSCRIPTEN_KEEPALIVE emscripten_async_run_in_main_thread(em_queued_call* call) {
-  do_emscripten_dispatch_to_thread(emscripten_main_browser_thread_id(), call);
+void emscripten_async_run_in_main_thread(em_queued_call* call) {
+  _emscripten_do_dispatch_to_thread(emscripten_main_browser_thread_id(), call);
 }
 
-void EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread(em_queued_call* call) {
+void emscripten_sync_run_in_main_thread(em_queued_call* call) {
   emscripten_async_run_in_main_thread(call);
 
   // Enter to wait for the operation to complete.
   emscripten_wait_for_call_v(call, INFINITY);
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_0(int function) {
+void* emscripten_sync_run_in_main_thread_0(int function) {
   em_queued_call q = {function};
   q.returnValue.vp = 0;
   emscripten_sync_run_in_main_thread(&q);
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_1(int function, void* arg1) {
+void* emscripten_sync_run_in_main_thread_1(int function, void* arg1) {
   em_queued_call q = {function};
   q.args[0].vp = arg1;
   q.returnValue.vp = 0;
@@ -555,7 +557,7 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_1(int function, vo
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_2(
+void* emscripten_sync_run_in_main_thread_2(
   int function, void* arg1, void* arg2) {
   em_queued_call q = {function};
   q.args[0].vp = arg1;
@@ -565,7 +567,7 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_2(
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_xprintf_varargs(
+void* emscripten_sync_run_in_main_thread_xprintf_varargs(
   int function, int param0, const char* format, ...) {
   va_list args;
   va_start(args, format);
@@ -588,7 +590,7 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_xprintf_varargs(
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_3(
+void* emscripten_sync_run_in_main_thread_3(
   int function, void* arg1, void* arg2, void* arg3) {
   em_queued_call q = {function};
   q.args[0].vp = arg1;
@@ -599,7 +601,7 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_3(
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_4(
+void* emscripten_sync_run_in_main_thread_4(
   int function, void* arg1, void* arg2, void* arg3, void* arg4) {
   em_queued_call q = {function};
   q.args[0].vp = arg1;
@@ -611,7 +613,7 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_4(
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_5(
+void* emscripten_sync_run_in_main_thread_5(
   int function, void* arg1, void* arg2, void* arg3, void* arg4, void* arg5) {
   em_queued_call q = {function};
   q.args[0].vp = arg1;
@@ -624,7 +626,7 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_5(
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_6(
+void* emscripten_sync_run_in_main_thread_6(
   int function, void* arg1, void* arg2, void* arg3, void* arg4, void* arg5, void* arg6) {
   em_queued_call q = {function};
   q.args[0].vp = arg1;
@@ -638,7 +640,7 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_6(
   return q.returnValue.vp;
 }
 
-void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_7(int function, void* arg1,
+void* emscripten_sync_run_in_main_thread_7(int function, void* arg1,
   void* arg2, void* arg3, void* arg4, void* arg5, void* arg6, void* arg7) {
   em_queued_call q = {function};
   q.args[0].vp = arg1;
@@ -653,40 +655,34 @@ void* EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_7(int function, vo
   return q.returnValue.vp;
 }
 
-void EMSCRIPTEN_KEEPALIVE emscripten_current_thread_process_queued_calls() {
+void emscripten_current_thread_process_queued_calls() {
   // #if PTHREADS_DEBUG == 2
-  //	EM_ASM(console.error('thread ' + _pthread_self() + ':
+  //  EM_ASM(console.error('thread ' + _pthread_self() + ':
   //emscripten_current_thread_process_queued_calls(), ' + new Error().stack));
   // #endif
 
-  // TODO: Under certain conditions we may want to have a nesting guard also for pthreads (and it
-  // will certainly be cleaner that way), but we don't yet have TLS variables outside
-  // pthread_set/getspecific, so convert this to TLS after TLS is implemented.
-  static int bool_main_thread_inside_nested_process_queued_calls = 0;
+  static thread_local bool thread_is_processing_queued_calls = false;
 
-  if (emscripten_is_main_browser_thread()) {
-    // It is possible that when processing a queued call, the call flow leads back to calling this
-    // function in a nested fashion! Therefore this scenario must explicitly be detected, and
-    // processing the queue must be avoided if we are nesting, or otherwise the same queued calls
-    // would be processed again and again.
-    if (bool_main_thread_inside_nested_process_queued_calls)
-      return;
-    // This must be before pthread_mutex_lock(), since pthread_mutex_lock() can call back to this
-    // function.
-    bool_main_thread_inside_nested_process_queued_calls = 1;
-  }
+  // It is possible that when processing a queued call, the control flow leads back to calling this
+  // function in a nested fashion! Therefore this scenario must explicitly be detected, and
+  // processing the queue must be avoided if we are nesting, or otherwise the same queued calls
+  // would be processed again and again.
+  if (thread_is_processing_queued_calls)
+    return;
+  // This must be before pthread_mutex_lock(), since pthread_mutex_lock() can call back to this
+  // function.
+  thread_is_processing_queued_calls = true;
 
   pthread_mutex_lock(&call_queue_lock);
   CallQueue* q = GetQueue(pthread_self());
   if (!q) {
     pthread_mutex_unlock(&call_queue_lock);
-    if (emscripten_is_main_browser_thread())
-      bool_main_thread_inside_nested_process_queued_calls = 0;
+    thread_is_processing_queued_calls = false;
     return;
   }
 
-  int head = emscripten_atomic_load_u32((void*)&q->call_queue_head);
-  int tail = emscripten_atomic_load_u32((void*)&q->call_queue_tail);
+  int head = q->call_queue_head;
+  int tail = q->call_queue_tail;
   while (head != tail) {
     // Assume that the call is heavy, so unlock access to the call queue while it is being
     // performed.
@@ -695,20 +691,24 @@ void EMSCRIPTEN_KEEPALIVE emscripten_current_thread_process_queued_calls() {
     pthread_mutex_lock(&call_queue_lock);
 
     head = (head + 1) % CALL_QUEUE_SIZE;
-    emscripten_atomic_store_u32((void*)&q->call_queue_head, head);
-    tail = emscripten_atomic_load_u32((void*)&q->call_queue_tail);
+    q->call_queue_head = head;
+    tail = q->call_queue_tail;
   }
   pthread_mutex_unlock(&call_queue_lock);
 
   // If the queue was full and we had waiters pending to get to put data to queue, wake them up.
   emscripten_futex_wake((void*)&q->call_queue_head, 0x7FFFFFFF);
 
-  if (emscripten_is_main_browser_thread())
-    bool_main_thread_inside_nested_process_queued_calls = 0;
+  thread_is_processing_queued_calls = false;
 }
 
-void EMSCRIPTEN_KEEPALIVE emscripten_main_thread_process_queued_calls() {
-  if (!emscripten_is_main_runtime_thread())
+// At times when we disallow the main thread to process queued calls, this will
+// be set to 0.
+int _emscripten_allow_main_runtime_queued_calls = 1;
+
+void emscripten_main_thread_process_queued_calls() {
+  if (!emscripten_is_main_runtime_thread() ||
+      !_emscripten_allow_main_runtime_queued_calls)
     return;
 
   emscripten_current_thread_process_queued_calls();
@@ -743,7 +743,7 @@ int emscripten_sync_run_in_main_runtime_thread_(EM_FUNC_SIGNATURE sig, void* fun
   return q.returnValue.i;
 }
 
-EMSCRIPTEN_KEEPALIVE double emscripten_run_in_main_runtime_thread_js(int index, int num_args, double* buffer, int sync) {
+double emscripten_run_in_main_runtime_thread_js(int index, int num_args, int64_t* buffer, int sync) {
   em_queued_call q;
   em_queued_call *c;
   if (sync) {
@@ -756,16 +756,19 @@ EMSCRIPTEN_KEEPALIVE double emscripten_run_in_main_runtime_thread_js(int index,
   c->calleeDelete = 1-sync;
   c->functionEnum = EM_PROXIED_JS_FUNCTION;
   c->functionPtr = (void*)index;
-  // We write out the JS doubles into args[], which must be of appropriate size - JS will assume that.
-  assert(sizeof(em_variant_val) == sizeof(double));
   assert(num_args+1 <= EM_QUEUED_JS_CALL_MAX_ARGS);
+  // The types are only known at runtime in these calls, so we store values that
+  // must be able to contain any valid JS value, including a 64-bit BigInt if
+  // BigInt support is enabled. We store to an i64, which can contain both a
+  // BigInt and a JS Number which is a 64-bit double.
   c->args[0].i = num_args;
   for (int i = 0; i < num_args; i++) {
-    c->args[i+1].d = buffer[i];
+    c->args[i+1].i64 = buffer[i];
   }
 
   if (sync) {
     emscripten_sync_run_in_main_thread(&q);
+    // TODO: support BigInt return values somehow.
     return q.returnValue.d;
   } else {
     // 'async' runs are fire and forget, where the caller detaches itself from the call object after
@@ -850,7 +853,7 @@ em_queued_call* emscripten_async_waitable_run_in_main_runtime_thread_(
   return q;
 }
 
-int EMSCRIPTEN_KEEPALIVE _emscripten_call_on_thread(
+int _emscripten_call_on_thread(
   int forceAsync,
   pthread_t targetThread, EM_FUNC_SIGNATURE sig, void* func_ptr, void* satellite, ...) {
   int numArguments = EM_FUNC_SIG_NUM_FUNC_ARGUMENTS(sig);
@@ -900,66 +903,46 @@ int EMSCRIPTEN_KEEPALIVE _emscripten_call_on_thread(
   if (forceAsync) {
     EM_ASM({
       setTimeout(function() {
-        _do_emscripten_dispatch_to_thread($0, $1);
+        __emscripten_do_dispatch_to_thread($0, $1);
       }, 0);
     }, targetThread, q);
     return 0;
   } else {
-    return do_emscripten_dispatch_to_thread(targetThread, q);
+    return _emscripten_do_dispatch_to_thread(targetThread, q);
   }
 }
 
-void llvm_memory_barrier() { emscripten_atomic_fence(); }
-
-int llvm_atomic_load_add_i32_p0i32(int* ptr, int delta) {
-  return emscripten_atomic_add_u32(ptr, delta);
-}
-
 // Stores the memory address that the main thread is waiting on, if any. If
 // the main thread is waiting, we wake it up before waking up any workers.
-EMSCRIPTEN_KEEPALIVE
-void* main_thread_futex;
+EMSCRIPTEN_KEEPALIVE void* _emscripten_main_thread_futex;
 
-typedef struct main_args {
-  int argc;
-  char** argv;
-} main_args;
+static int _main_argc;
+static char** _main_argv;
 
 extern int __call_main(int argc, char** argv);
 
-void* __emscripten_thread_main(void* param) {
-  emscripten_set_thread_name(pthread_self(),
-    "Application main thread"); // This is the main runtime thread for the application.
-  main_args* args = (main_args*)param;
-  return (void*)__call_main(args->argc, args->argv);
+static void* _main_thread(void* param) {
+  // This is the main runtime thread for the application.
+  emscripten_set_thread_name(pthread_self(), "Application main thread");
+  return (void*)__call_main(_main_argc, _main_argv);
 }
 
-static main_args _main_arguments;
-
-int proxy_main(int argc, char** argv) {
-  if (emscripten_has_threading_support()) {
-    pthread_attr_t attr;
-    pthread_attr_init(&attr);
-    // Use the size of the current stack, which is the normal size of the stack
-    // that main() would have without PROXY_TO_PTHREAD.
-    pthread_attr_setstacksize(&attr, emscripten_stack_get_base() - emscripten_stack_get_end());
-    // Pass special ID -1 to the list of transferred canvases to denote that the thread creation
-    // should instead take a list of canvases that are specified from the command line with
-    // -s OFFSCREENCANVASES_TO_PTHREAD linker flag.
-    emscripten_pthread_attr_settransferredcanvases(&attr, (const char*)-1);
-    _main_arguments.argc = argc;
-    _main_arguments.argv = argv;
-    pthread_t thread;
-    int rc = pthread_create(&thread, &attr, __emscripten_thread_main, (void*)&_main_arguments);
-    pthread_attr_destroy(&attr);
-    if (rc) {
-      // Proceed by running main() on the main browser thread as a fallback.
-      return __call_main(_main_arguments.argc, _main_arguments.argv);
-    }
-    return 0;
-  } else {
-    return __call_main(_main_arguments.argc, _main_arguments.argv);
-  }
+int emscripten_proxy_main(int argc, char** argv) {
+  pthread_attr_t attr;
+  pthread_attr_init(&attr);
+  // Use the size of the current stack, which is the normal size of the stack
+  // that main() would have without PROXY_TO_PTHREAD.
+  pthread_attr_setstacksize(&attr, emscripten_stack_get_base() - emscripten_stack_get_end());
+  // Pass special ID -1 to the list of transferred canvases to denote that the thread creation
+  // should instead take a list of canvases that are specified from the command line with
+  // -s OFFSCREENCANVASES_TO_PTHREAD linker flag.
+  emscripten_pthread_attr_settransferredcanvases(&attr, (const char*)-1);
+  _main_argc = argc;
+  _main_argv = argv;
+  pthread_t thread;
+  int rc = pthread_create(&thread, &attr, _main_thread, NULL);
+  pthread_attr_destroy(&attr);
+  return rc;
 }
 
 weak_alias(__pthread_testcancel, pthread_testcancel);
