@@ -27,7 +27,6 @@ import difflib
 import fnmatch
 import glob
 import hashlib
-import json
 import logging
 import math
 import multiprocessing
@@ -44,6 +43,7 @@ import tempfile
 import time
 import unittest
 import webbrowser
+from pathlib import Path
 from http.server import HTTPServer, SimpleHTTPRequestHandler
 from urllib.parse import unquote, unquote_plus
 
@@ -66,7 +66,7 @@ from tools import shared, line_endings, building, config
 
 def path_from_root(*pathelems):
   """Construct a path relative to the emscripten root directory."""
-  return os.path.join(__rootpath__, *pathelems)
+  return str(Path(__rootpath__, *pathelems))
 
 
 sys.path.append(path_from_root('third_party/websockify'))
@@ -98,7 +98,9 @@ EMTEST_VERBOSE = int(os.getenv('EMTEST_VERBOSE', '0')) or shared.DEBUG
 
 TEST_ROOT = path_from_root('tests')
 
-WEBIDL_BINDER = shared.bat_suffix(path_from_root('tools', 'webidl_binder'))
+WEBIDL_BINDER = shared.bat_suffix(path_from_root('tools/webidl_binder'))
+
+EMBUILDER = shared.bat_suffix(path_from_root('embuilder'))
 
 
 if EMTEST_VERBOSE:
@@ -112,7 +114,15 @@ def delete_contents(pathname):
 
 def test_file(*path_components):
   """Construct a path relative to the emscripten "tests" directory."""
-  return os.path.join(TEST_ROOT, *path_components)
+  return str(Path(TEST_ROOT, *path_components))
+
+
+def read_file(*path_components):
+  return Path(*path_components).read_text()
+
+
+def read_binary(*path_components):
+  return Path(*path_components).read_bytes()
 
 
 # checks if browser testing is enabled
@@ -120,6 +130,13 @@ def has_browser():
   return EMTEST_BROWSER != '0'
 
 
+def compiler_for(filename, force_c=False):
+  if shared.suffix(filename) in ('.cc', '.cxx', '.cpp') and not force_c:
+    return EMXX
+  else:
+    return EMCC
+
+
 # Generic decorator that calls a function named 'condition' on the test class and
 # skips the test if that function returns true
 def skip_if(func, condition, explanation='', negate=False):
@@ -191,6 +208,26 @@ def requires_native_clang(func):
   return decorated
 
 
+def require_node(func):
+  assert callable(func)
+
+  def decorated(self, *args, **kwargs):
+    self.require_node()
+    return func(self, *args, **kwargs)
+
+  return decorated
+
+
+def require_v8(func):
+  assert callable(func)
+
+  def decorated(self, *args, **kwargs):
+    self.require_v8()
+    return func(self, *args, **kwargs)
+
+  return decorated
+
+
 def node_pthreads(f):
   def decorated(self):
     self.set_setting('USE_PTHREADS')
@@ -235,8 +272,8 @@ def with_env_modify(updates):
 
 
 def ensure_dir(dirname):
-  if not os.path.isdir(dirname):
-    os.makedirs(dirname)
+  dirname = Path(dirname)
+  dirname.mkdir(parents=True, exist_ok=True)
 
 
 def limit_size(string, maxbytes=800000 * 20, maxlines=100000, max_line=5000):
@@ -253,14 +290,16 @@ def limit_size(string, maxbytes=800000 * 20, maxlines=100000, max_line=5000):
 
 
 def create_file(name, contents, binary=False):
-  assert not os.path.isabs(name)
-  mode = 'wb' if binary else 'w'
-  with open(name, mode) as f:
-    f.write(contents)
+  name = Path(name)
+  assert not name.is_absolute()
+  if binary:
+    name.write_bytes(contents)
+  else:
+    name.write_text(contents)
 
 
 def make_executable(name):
-  os.chmod(name, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)
+  Path(name).chmod(stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)
 
 
 # The core test modes
@@ -396,7 +435,25 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
     if not self.is_wasm():
       self.skipTest('no dynamic linking support in wasm2js yet')
     if '-fsanitize=address' in self.emcc_args:
-      self.skipTest('no dynamic linking support in asan yet')
+      self.skipTest('no dynamic linking support in ASan yet')
+    if '-fsanitize=leak' in self.emcc_args:
+      self.skipTest('no dynamic linking support in LSan yet')
+
+  def require_v8(self):
+    if not config.V8_ENGINE or config.V8_ENGINE not in config.JS_ENGINES:
+      if 'EMTEST_SKIP_V8' in os.environ:
+        self.skipTest('test requires v8 and EMTEST_SKIP_V8 is set')
+      else:
+        self.fail('d8 required to run this test.  Use EMTEST_SKIP_V8 to skip')
+    self.js_engines = [config.V8_ENGINE]
+
+  def require_node(self):
+    if not config.NODE_JS or config.NODE_JS not in config.JS_ENGINES:
+      if 'EMTEST_SKIP_NODE' in os.environ:
+        self.skipTest('test requires node and EMTEST_SKIP_NODE is set')
+      else:
+        self.fail('node required to run this test.  Use EMTEST_SKIP_NODE to skip')
+    self.js_engines = [config.NODE_JS]
 
   def uses_memory_init_file(self):
     if self.get_setting('SIDE_MODULE') or (self.is_wasm() and not self.get_setting('WASM2JS')):
@@ -416,12 +473,12 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
 
   @classmethod
   def setUpClass(cls):
-    super(RunnerCore, cls).setUpClass()
+    super().setUpClass()
     print('(checking sanity from test runner)') # do this after we set env stuff
     shared.check_sanity(force=True)
 
   def setUp(self):
-    super(RunnerCore, self).setUp()
+    super().setUp()
     self.settings_mods = {}
     self.emcc_args = ['-Werror']
     self.node_args = []
@@ -516,11 +573,11 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
     ret = []
     for key, value in self.settings_mods.items():
       if value == 1:
-        ret += ['-s', key]
-      elif type(value) == str:
-        ret += ['-s', f'{key}={value}']
+        ret.append(f'-s{key}')
+      elif type(value) == list:
+        ret.append(f'-s{key}={",".join(value)}')
       else:
-        ret += ['-s', f'{key}={json.dumps(value)}']
+        ret.append(f'-s{key}={value}')
     return ret
 
   def get_dir(self):
@@ -569,26 +626,25 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
 
   # Build JavaScript code from source code
   def build(self, filename, libraries=[], includes=[], force_c=False,
-            post_build=None, js_outfile=True):
+            post_build=None, js_outfile=True, emcc_args=[]):
     suffix = '.js' if js_outfile else '.wasm'
-    if shared.suffix(filename) in ('.cc', '.cxx', '.cpp') and not force_c:
-      compiler = [EMXX]
-    else:
+    compiler = [compiler_for(filename, force_c)]
+    if compiler[0] == EMCC:
       # TODO(https://github.com/emscripten-core/emscripten/issues/11121)
       # We link with C++ stdlibs, even when linking with emcc for historical reasons.  We can remove
       # this if this issues is fixed.
-      compiler = [EMCC, '-nostdlib++']
+      compiler.append('-nostdlib++')
 
     if force_c:
       compiler.append('-xc')
 
     dirname, basename = os.path.split(filename)
     output = shared.unsuffixed(basename) + suffix
-    cmd = compiler + [filename, '-o', output] + self.get_emcc_args(main_file=True) + libraries
+    cmd = compiler + [filename, '-o', output] + self.get_emcc_args(main_file=True) + emcc_args + libraries
     if shared.suffix(filename) not in ('.i', '.ii'):
       # Add the location of the test file to include path.
       cmd += ['-I.']
-      cmd += ['-I' + include for include in includes]
+      cmd += ['-I' + str(include) for include in includes]
 
     self.run_process(cmd, stderr=self.stderr_redirect if not DEBUG else None)
     self.assertExists(output)
@@ -599,7 +655,7 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
       post_build(output)
 
     if js_outfile and self.uses_memory_init_file():
-      src = open(output).read()
+      src = read_file(output)
       # side memory init file, or an empty one in the js
       assert ('/* memory initializer */' not in src) or ('/* memory initializer */ allocate([]' in src)
 
@@ -626,8 +682,8 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
     start_off = 0
     end_off = 0
 
-    with open(javascript_file, 'rt') as f:
-      blob = "".join(f.readlines())
+    js = read_file(javascript_file)
+    blob = "".join(js.splitlines())
 
     start_off = blob.find(start_tok) + len(start_tok)
     end_off = blob.find(end_tok)
@@ -658,7 +714,7 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
     stderr = self.in_dir('stderr')
     error = None
     if not engine:
-      engine = config.JS_ENGINES[0]
+      engine = self.js_engines[0]
     if engine == config.NODE_JS:
       engine = engine + self.node_args
     if engine == config.V8_ENGINE:
@@ -677,8 +733,8 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
     if not filename.endswith('.wasm'):
       self.assertEqual(line_endings.check_line_endings(filename), 0)
 
-    out = open(stdout, 'r').read()
-    err = open(stderr, 'r').read()
+    out = read_file(stdout)
+    err = read_file(stderr)
     if output_nicerizer:
       ret = output_nicerizer(out, err)
     else:
@@ -780,8 +836,8 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
   def assertBinaryEqual(self, file1, file2):
     self.assertEqual(os.path.getsize(file1),
                      os.path.getsize(file2))
-    self.assertEqual(open(file1, 'rb').read(),
-                     open(file2, 'rb').read())
+    self.assertEqual(read_binary(file1),
+                     read_binary(file2))
 
   library_cache = {}
 
@@ -843,6 +899,12 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
         self.fail('subprocess exited with non-zero return code(%d): `%s`' %
                   (e.returncode, shared.shlex_join(cmd)))
 
+  def emcc(self, filename, args=[], output_filename=None, **kwargs):
+    if output_filename is None:
+      output_filename = filename + '.o'
+    try_delete(output_filename)
+    self.run_process([compiler_for(filename), filename] + args + ['-o', output_filename], **kwargs)
+
   # Shared test code between main suite and others
 
   def expect_fail(self, cmd, **args):
@@ -896,28 +958,22 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
         static ainit _;
       ''')
 
-    create_file('libb.cpp', r'''
+    create_file('libb.c', r'''
         #include <emscripten.h>
 
-        extern "C" {
         void afunc(const char *s);
-        EMSCRIPTEN_KEEPALIVE void bfunc();
-        }
 
-        void bfunc() {
+        EMSCRIPTEN_KEEPALIVE void bfunc() {
           afunc("b");
         }
       ''')
 
-    create_file('libc.cpp', r'''
+    create_file('libc.c', r'''
         #include <emscripten.h>
 
-        extern "C" {
         void afunc(const char *s);
-        EMSCRIPTEN_KEEPALIVE void cfunc();
-        }
 
-        void cfunc() {
+        EMSCRIPTEN_KEEPALIVE void cfunc() {
           afunc("c");
         }
       ''')
@@ -938,13 +994,11 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
       self.run_process(cmdv)
 
     ccshared('liba.cpp')
-    ccshared('libb.cpp', ['liba' + so])
-    ccshared('libc.cpp', ['liba' + so])
+    ccshared('libb.c', ['liba' + so])
+    ccshared('libc.c', ['liba' + so])
 
     self.set_setting('MAIN_MODULE')
-    original_args = self.emcc_args.copy()
-    extra_args = ['libb' + so, 'libc' + so]
-    self.emcc_args += extra_args
+    extra_args = ['-L.', 'libb' + so, 'libc' + so]
     do_run(r'''
       #ifdef __cplusplus
       extern "C" {
@@ -961,9 +1015,8 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
         return 0;
       }
       ''',
-           'a: loaded\na: b (prev: (null))\na: c (prev: b)\n')
+           'a: loaded\na: b (prev: (null))\na: c (prev: b)\n', emcc_args=extra_args)
 
-    self.emcc_args = original_args
     for libname in ['liba', 'libb', 'libc']:
       self.emcc_args += ['--embed-file', libname + so]
     do_run(r'''
@@ -973,7 +1026,7 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
 
       int test_main() {
         void *bdso, *cdso;
-        void (*bfunc)(), (*cfunc)();
+        void (*bfunc_ptr)(), (*cfunc_ptr)();
 
         // FIXME for RTLD_LOCAL binding symbols to loaded lib is not currently working
         bdso = dlopen("libb%(so)s", RTLD_NOW|RTLD_GLOBAL);
@@ -981,13 +1034,13 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
         cdso = dlopen("libc%(so)s", RTLD_NOW|RTLD_GLOBAL);
         assert(cdso != NULL);
 
-        bfunc = (void (*)())dlsym(bdso, "bfunc");
-        assert(bfunc != NULL);
-        cfunc = (void (*)())dlsym(cdso, "cfunc");
-        assert(cfunc != NULL);
+        bfunc_ptr = (void (*)())dlsym(bdso, "bfunc");
+        assert(bfunc_ptr != NULL);
+        cfunc_ptr = (void (*)())dlsym(cdso, "cfunc");
+        assert(cfunc_ptr != NULL);
 
-        bfunc();
-        cfunc();
+        bfunc_ptr();
+        cfunc_ptr();
         return 0;
       }
     ''' % locals(),
@@ -1021,12 +1074,13 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
 
   ## Just like `do_run` but with filename of expected output
   def do_run_from_file(self, filename, expected_output_filename, **kwargs):
-    self._build_and_run(filename, open(expected_output_filename).read(), **kwargs)
+    self._build_and_run(filename, read_file(expected_output_filename), **kwargs)
 
   def do_run_in_out_file_test(self, *path, **kwargs):
     srcfile = test_file(*path)
-    outfile = shared.unsuffixed(srcfile) + '.out'
-    expected = open(outfile).read()
+    out_suffix = kwargs.pop('out_suffix', '')
+    outfile = shared.unsuffixed(srcfile) + out_suffix + '.out'
+    expected = read_file(outfile)
     self._build_and_run(srcfile, expected, **kwargs)
 
   ## Does a complete test - builds, runs, checks output, etc.
@@ -1035,14 +1089,14 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
                      js_engines=None, post_build=None, libraries=[],
                      includes=[],
                      assert_returncode=0, assert_identical=False, assert_all=False,
-                     check_for_error=True, force_c=False):
+                     check_for_error=True, force_c=False, emcc_args=[]):
     logger.debug(f'_build_and_run: {filename}')
 
     if no_build:
       js_file = filename
     else:
       self.build(filename, libraries=libraries, includes=includes, post_build=post_build,
-                 force_c=force_c)
+                 force_c=force_c, emcc_args=emcc_args)
       js_file = shared.unsuffixed(os.path.basename(filename)) + '.js'
     self.assertExists(js_file)
 
@@ -1096,8 +1150,8 @@ class RunnerCore(unittest.TestCase, metaclass=RunnerMeta):
     self.set_setting('ERROR_ON_UNDEFINED_SYMBOLS', 0)
 
     self.emcc_args += [
-      '-I' + test_file('third_party', 'freetype', 'include'),
-      '-I' + test_file('third_party', 'poppler', 'include')
+      '-I' + test_file('third_party/freetype/include'),
+      '-I' + test_file('third_party/poppler/include')
     ]
 
     freetype = self.get_freetype_library()
@@ -1173,7 +1227,7 @@ def harness_server_func(in_queue, out_queue, port):
         self.send_response(200)
         self.send_header('Content-type', 'text/html')
         self.end_headers()
-        self.wfile.write(open(test_file('browser_harness.html'), 'rb').read())
+        self.wfile.write(read_binary(test_file('browser_harness.html')))
       elif 'report_' in self.path:
         # the test is reporting its result. first change dir away from the
         # test dir, as it will be deleted now that the test is finishing, and
@@ -1277,7 +1331,7 @@ class BrowserCore(RunnerCore):
   unresponsive_tests = 0
 
   def __init__(self, *args, **kwargs):
-    super(BrowserCore, self).__init__(*args, **kwargs)
+    super().__init__(*args, **kwargs)
 
   @staticmethod
   def browser_open(url):
@@ -1305,7 +1359,7 @@ class BrowserCore(RunnerCore):
 
   @classmethod
   def setUpClass(cls):
-    super(BrowserCore, cls).setUpClass()
+    super().setUpClass()
     cls.also_asmjs = int(os.getenv('EMTEST_BROWSER_ALSO_ASMJS', '0')) == 1
     cls.port = int(os.getenv('EMTEST_BROWSER_PORT', '8888'))
     if not has_browser():
@@ -1320,7 +1374,7 @@ class BrowserCore(RunnerCore):
 
   @classmethod
   def tearDownClass(cls):
-    super(BrowserCore, cls).tearDownClass()
+    super().tearDownClass()
     if not has_browser():
       return
     cls.harness_server.terminate()
@@ -1405,9 +1459,9 @@ class BrowserCore(RunnerCore):
     #   pngcrush -rem gAMA -rem cHRM -rem iCCP -rem sRGB infile outfile
     basename = os.path.basename(expected)
     shutil.copyfile(expected, os.path.join(self.get_dir(), basename))
+    reporting = read_file(test_file('browser_reporting.js'))
     with open('reftest.js', 'w') as out:
-      with open(test_file('browser_reporting.js')) as reporting:
-        out.write('''
+      out.write('''
       function doReftest() {
         if (doReftest.done) return;
         doReftest.done = true;
@@ -1503,7 +1557,7 @@ class BrowserCore(RunnerCore):
           setTimeout(realDoReftest, 1);
         };
       }
-''' % (reporting.read(), basename, int(manually_trigger)))
+''' % (reporting, basename, int(manually_trigger)))
 
   def compile_btest(self, args, reporting=Reporting.FULL):
     # Inject support code for reporting results. This adds an include a header so testcases can
@@ -1609,10 +1663,11 @@ def build_library(name,
     generated_libs = [generated_libs]
   source_dir = test_file(name.replace('_native', ''))
 
-  project_dir = os.path.join(build_dir, name)
+  project_dir = Path(build_dir, name)
   if os.path.exists(project_dir):
     shutil.rmtree(project_dir)
-  shutil.copytree(source_dir, project_dir) # Useful in debugging sometimes to comment this out, and two lines above
+  # Useful in debugging sometimes to comment this out, and two lines above
+  shutil.copytree(source_dir, project_dir)
 
   generated_libs = [os.path.join(project_dir, lib) for lib in generated_libs]
   if native:
@@ -1634,14 +1689,12 @@ def build_library(name,
           shared.run_process(configure, env=env, stdout=stdout, stderr=stderr,
                              cwd=project_dir)
     except subprocess.CalledProcessError:
-      with open(os.path.join(project_dir, 'configure_out')) as f:
-        print('-- configure stdout --')
-        print(f.read())
-        print('-- end configure stdout --')
-      with open(os.path.join(project_dir, 'configure_err')) as f:
-        print('-- configure stderr --')
-        print(f.read())
-        print('-- end configure stderr --')
+      print('-- configure stdout --')
+      print(read_file(Path(project_dir, 'configure_out')))
+      print('-- end configure stdout --')
+      print('-- configure stderr --')
+      print(read_file(Path(project_dir, 'configure_err')))
+      print('-- end configure stderr --')
       raise
 
   def open_make_out(mode='r'):
@@ -1675,7 +1728,7 @@ def build_library(name,
     cache[cache_name] = []
     for f in generated_libs:
       basename = os.path.basename(f)
-      cache[cache_name].append((basename, open(f, 'rb').read()))
+      cache[cache_name].append((basename, read_binary(f)))
 
   return generated_libs
 
